#! /usr/bin/python3

'''
 VERSION 20200727
'''

import argparse

def user_input():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', help='File containing the coarse-grained structure of the protein in pdb format.',required=True)
    parser.add_argument('-f', help='File containing the contact analysis of the (atomistic) protein structure obtained from the webserver http://info.ifpan.edu.pl/~rcsu/rcsu/index.html.',required=True)
    parser.add_argument('--go_eps', type=float, default=9.414, help='Dissociation energy [kJ/mol] of the Lennard-Jones potential used in the Go-like model (default: 9.414).')
    parser.add_argument('--cutoff_short', type=float, default=0.3,   help='Lower cutoff distance [nm]: contacts with a shorter distance than cutoff_short are not included in the Go-like interactions (default: 0.3).')
    parser.add_argument('--cutoff_long',  type=float, default=1.1,   help='Upper cutoff distance [nm]: contacts with a longer distance than cutoff_long are not included in the Go-like interactions (default: 1.1).')
    parser.add_argument('--symm', default=True, help='Select only symmetrical contacts.',action='store_true')
    #parser.add_argument('--go_dist', type=float, default=4  ,   help='Minimum chain distance among two contacts.')
    #parser.add_argument('--no-OV', default=False, help='Do not use OV contact map.',action='store_true')
    #parser.add_argument('--no-rCSU', default=False, help='Do not use rCSU contact map',action='store_true')
    #parser.add_argument('--no_interface',default=False, help='Do not write Go interactions for different molecules',action='store_true')
    parser.add_argument('--verbose', default=False, help='Print all decisions made about the contacts.',action='store_true')
    return parser.parse_args()

def read_coarse_grain_pdbs(files):
    '''
    INPUT: COMMA SEPARATED LIST OF MARTINI COARSE-GRAINED PDB FILES
    OUTPUT: LIST OF (GLOBAL) BB COORDINATES
    '''
    pdb = []
    def readpdb(name):
        name = name.strip()
        with open(name,'r') as fid:
            for line in fid:
                if line.find('ATOM') == 0 or line.find('HETATM') == 0:
                    aname = line[12:16].strip() # Atom name
                    resname = line[17:20].strip() # Residue name
                    resseq = int(line[22:26])    # Residue sequence number
                    x,y,z = float(line[30:38]),float(line[38:46]),float(line[46:54]) # coordinates [x,y,z]
                    pdb[-1].append([resseq,resname,aname,x,y,z])
        print('\tTotal number of coordinates read: ',len(pdb[-1]))
    print(files)
    for f in files.split(','):
         pdb.append([])
         print('Reading ATOM and HETATM from:',f)
         readpdb(f)
    return pdb

def read_contact_map(file_map):
    '''
    INPUT: CONTACT MAP FROM contact_map
    OUTPUT: GENERATOR OF PAIRS OF CONTACTS OV+rCSU
    '''
    print('Reading contact map file:',file_map)
    import re
    OV,rCSU = re.compile('1 [01] [01] [01]'),re.compile('[01] [01] [01] 1')
    nOV,nrCSU = 0,0
    def map(line):
         #                      I1              I2
         idx1,idx2 = int(line[10:16]),int(line[29:35])
         return [idx1,idx2]
    with open(file_map,'r') as fid:
         for line in fid:
             ov = OV.search(line)
             rcsu = rCSU.search(line)
             if (ov   != None): nOV   = nOV+1
             if (rcsu != None): nrCSU = nrCSU+1
             if (ov != None or rcsu != None):
                 yield(map(line))
    print('\tNumber of   OV  contacts read: ', nOV)
    print('\tNumber of  rSCU contacts read: ', nrCSU)
    print('\tTotal number of contacts read: ', nOV+nrCSU)

def get_goMARTINI_contacts(args, dataBB, file_map, cutoff_short, cutoff_long, go_eps):
    VERBOSE = args.verbose
    SYMM = args.symm
    print('Selecting GoMARTINI contacts')
    # parameters of GoMARTINI model
    cutoff_short, cutoff_long, go_eps = float(cutoff_short), float(cutoff_long), float(go_eps)
    seqDist = 4
    def distance(c1,c2):
        import math
        dx,dy,dz=c1[0]-c2[0],c1[1]-c2[1],c1[2]-c2[2]
        return math.sqrt(dx*dx + dy*dy + dz*dz)
    def select_valid_go_contacts():
        ncontacts = 0
        buffer = {}
        if VERBOSE: print('Contacts excluded:')
        for contact in read_contact_map(file_map):
            ncontacts = ncontacts+1
            idx0,idx1 = contact[0]-1,contact[1]-1
            coords0,coords1 = dataBB[idx0][1:],dataBB[idx1][1:]
            res0,res1 = dataBB[idx0][0],dataBB[idx1][0]
            distanceBB = distance(coords0,coords1)/10.0 # [Ang] to [nm]
            # select valid Go contacs
            excluded = False
            if (abs(res0 - res1) < seqDist):
                excluded = True
                if VERBOSE: print(ncontacts,'\tless than',seqDist-1,'residues:  %s %s %f'%(res0, res1, distanceBB))
            if distanceBB > cutoff_long:
                excluded = True
                if VERBOSE: print(ncontacts,'\tdistance > cutoff_long:  %s %s %f'%(res0, res1, distanceBB))
            if distanceBB < cutoff_short:
                excluded = True
                if VERBOSE: print(ncontacts,'\tdistance < cutoff_short: %s %s %f'%(res0, res1, distanceBB))
            if not excluded:
                id = '%d %d'%(idx0,idx1)
                buffer[id] = distanceBB
        # symmetric contacts
        ncontacts = 0
        for id0,dist in buffer.items():
            if dist:
                i0,i1 = id0.split()
                id1 = '%s %s'%(i1,i0)
                if SYMM:
                    if id1 in buffer:
                        ncontacts = ncontacts+1
                        buffer[id1] = False
                        buffer[id0] = False
                        yield(int(i0),int(i1),dist)
                    else:
                        buffer[id0] = False
                        if VERBOSE: print('\tnot a symmetric contact: %s'%(id0))
                else:
                    ncontacts = ncontacts+1
                    buffer[id1] = False
                    buffer[id0] = False
                    yield(int(i0),int(i1),dist)
        print('\tNumber of valid GoMARTINI contacts:', ncontacts)
        print("\t-----------------------------------")
        print('\tThese results exclude the contacts with distances higher than the cutoff_long (' + str(cutoff_long) + ' nm), shorter than the cutoff_short (' +
            str(cutoff_short) + ' nm), or where the AA have less than ' + str(seqDist-1) + ' other AA between each other')
        print("\t-----------------------------------")
    data = []
    for i0,i1,dist in select_valid_go_contacts():
        # parameters for LJ potential
        # calc sigma for the LJ potential in [nm]
        sigma = dist / 1.12246204830
        #Vii = 4.0 * pow(sigma,6) * go_eps
        #Wii = 4.0 * pow(sigma,12) * go_eps
        data.append([i0,i1,sigma])
        #print("%5d (%5d) %5d (%5d) %15.10f"%(i0,res0,i1,res1,sigma))
    return data

def parser_itp(file):
    name = "%s.itp"%file[:-4]
    print('\tParsing %s'%name)
    f = open(name,'r')
    d = {}
    key = "header"
    d[key] = []
    for l in f:
        line = l.strip()
        if ('[' in line) and (']' in line):
            key = line.replace(']','').replace('[','').strip()
            d[key] = []
        d[key].append(line)
    f.close()
    return d

if __name__ == '__main__':
    print(
"             CREATE GoMARTINI MODEL FROM MARTINI PDB/ITP FILES                \n"
"                                                                              \n"
" This software is written by:                                                 \n"
"       Rodrigo Azevedo Moreira da Silva                                       \n"
"                                                                              \n"
" Copyright (c) 2020 - IPPT-PAN                                                \n"
"       Institute of Fundamental Techonological Research                       \n"
"       Polish Academy of Sciences                                             \n"
" GNU AFFERO GENERAL PUBLIC LICENSE, check out LICENSE for more informations.  \n"
"                                                                              \n"
)
    args = user_input()

    import sys
    print(str(sys.argv))

    for f in args.s.split(','):
        name = f[:-4]
        try: open("%s.pdb"%name,'r')
        except:
            print("ERROR: coarse-grain MARTINI PDB %s.pdb not found!"%(name))
            exit(1)
        try: open("%s.itp"%name,'r')
        except:
            print("ERROR: coarse-grain MARTINI ITP %s.itp not found!"%(name))
            exit(1)

    pdb = read_coarse_grain_pdbs(args.s)

    gomartini = open('go_martini.itp','w')
    gomartini.write('\n[ atomtypes ]\n')
    dataBB = []
    names = []
    natoms = 0
    for c,p in enumerate(pdb):
        serial = 0
        for atom in p:
            natoms = natoms + 1
            if atom[2] == 'BB':
                serial = serial + 1
                s = "GO_%d_%d"%(c+1,serial)
                gomartini.write("%-10s 0.0 0.0 A 0.0 0.0\n"%(s))
                names.append(s)
                dataBB.append([atom[0],atom[3],atom[4],atom[5]])
    gomartini.write('\n[ nonbond_params ]\n')
    contacts = get_goMARTINI_contacts(args,dataBB, args.f, args.cutoff_short, args.cutoff_long, args.go_eps)
    for v in contacts:
        n1,n2,sigma = names[int(v[0])],names[int(v[1])],v[2]
        eps = args.go_eps
        Vii = 4*eps*(sigma**6)
        Wii = Vii*(sigma**6)
        # com rule 1
        # gomartini.write("%-10s %-10s 1 %14.10f %14.10f ; sigma %f epsilon %f\n"%(n1,n2,Vii,Wii,sigma,eps))
        # comb rule 2 or 3 (sigma, eps)
        gomartini.write("%-10s %-10s 1 %14.10f %14.10f\n"%(n1,n2,sigma,eps))
    gomartini.close()
    print('Wrote "go_martini.itp"')

    gro = open('go_system.gro','w')
    gro.write('system coordinates\n')
    gro.write('%d\n'%(natoms+len(names)))
    serial = 0
    for npdb,p in enumerate(pdb):
        for atom in p:
            serial = serial + 1
            gro.write("%5d%-5s%5s%5d%8.3f%8.3f%8.3f\n"%(atom[0],atom[1],atom[2],serial,atom[3]/10,atom[4]/10,atom[5]/10))
        for atom in p:
            if atom[2] == 'BB':
                serial = serial + 1
                gro.write("%5d%-5s%5s%5d%8.3f%8.3f%8.3f\n"%(atom[0],atom[1],'CA',serial,atom[3]/10,atom[4]/10,atom[5]/10))
    gro.write('0.000 0.000 0.000')
    gro.close()
    print('Wrote "go_system.gro"')

    serial = 0
    serial2 = 0
    for k in range(len(pdb)):
        itp = parser_itp(args.s.split(',')[k])
        mol = open('go_molecule%d.itp'%(k+1),'w')
        for l in itp["header"]: mol.write("%s\n"%l)
        mol.write("[ moleculetype ]\ngo_molecule%d 1\n\n"%(k+1))
        #for l in itp["moleculetype"]: mol.write("%s\n"%l)
        for l in itp["atoms"]:
            if len(l) > 0:
                mol.write("%s\n"%l)
        #mol.write('\n[ atoms ]\n')
        l = len(pdb[k])
        res = pdb[k][-1][0]
        for natom,atom in enumerate(pdb[k]):
            if atom[2] == 'BB':
                res = res + 1
                l = l+1
                serial2 = serial2 + 1
                mol.write("%d %-s %d %3s CA %d 0.0\n"%(l,names[serial2-1],atom[0],atom[1],l))
        mol.write('')
        for l in itp["bonds"]: mol.write("%s\n"%l)
        for l in itp["constraints"]: mol.write("%s\n"%l)
        for l in itp["angles"]: mol.write("%s\n"%l)
        if 'virtual_sitesn' in itp:
               for l in itp["virtual_sitesn"]: mol.write("%s\n"%l)
        else:
               mol.write('\n[ virtual_sitesn ]\n')
        l = len(pdb[k])
        res = {}
        for natom,atom in enumerate(pdb[k]):
            if atom[2] == 'BB':
                serial = serial + 1
                l = l+1
                mol.write("%d 1 %d\n"%(l,natom+1))
                res[serial] = natom+1
        mol.write('')
        if 'exclusions' in itp:
               for l in itp["exclusions"]: mol.write("%s\n"%l)
        else:
               mol.write('\n[ exclusions ]\n')
        for v in contacts:
            r1,r2 = v[0],v[1]
            if (r1 in res) and (r2 in res):
                mol.write("%-7d %-7d ; %10s %10s\n"%(res[r1],res[r2],names[r1-1],names[r2-1]))
        for l in itp["dihedrals"]: mol.write("%s\n"%l)
        for l in itp["position_restraints"]: mol.write("%s\n"%l)
        mol.write('')
        print('Wrote "go_molecule%d.itp"'%(k+1))
        mol.close()

    print("\n\n-----------------------------------")
    print("DEFAULT PARAMETERS")
    print("-----------------------------------")
    print('''
	; RUN CONTROL PARAMETERS
	integrator               = md
	dt                       = 0.02

	; OUTPUT CONTROL OPTIONS
	nstxout                  = 20000
	nstvout                  = 20000
	nstfout                  = 20000
	nstlog                   = 20000
	nstenergy                = 20000
	nstxout-compressed       = 20000
	compressed-x-precision   = 100

	; OPTIONS FOR ELECTROSTATICS AND VDW
	coulombtype              = reaction-field
	rcoulomb                 = 1.1
	epsilon_r                = 15	; 2.5 (with polarizable water)
	epsilon_rf               = 0
	vdw_type                 = cutoff
	vdw-modifier             = Potential-shift-verlet
	rvdw                     = 1.1
''')
    print("-----------------------------------")
    print("TOPOLOGY FILE")
    print("-----------------------------------\n\n")
    print('#include "martini_v3.0.4.itp"')
    print('#include "go_martini.itp"')
    print('#include "martini_v3.0_ions.itp"')
    print('#include "martini_v3.0_solvents.itp"')
    c = 0
    for f in args.s.split(','):
        c = c+1
        #print('\n#include "%s.itp"'%f[:-4])
        print('#include "go_molecule%d.itp"'%(c))
    print('\n[ system ]')
    print('SYSTEM NAME')
    print('\n[ molecules ]')
    c = 0
    for f in args.s.split(','):
        c = c+1
        print("go_molecule%d\t1"%(c))
        #print('%s    1'%f[:-4])
 
